"use strict";
this.name           = "NShields";
this.author         = "Ngalo";
this.license        = "CC-BY-NC-SA 3.0";
this.description    = "Main script for N-Shields NPC shield simulation and visual effects";
this.version        = "1.0";

//==== SETTINGS ===========================================================================

this._NPC_shields = true;
this._Thargoid_shields = true;
this._Thargoid_energy_reduction = true;

this._drawFlashers = true;
this._drawExploders = false;
this._drawSparkEffects = true;

this._NPC_flasherColours = 1; //multicolour, type 2
this._Thargoid_flasherColours = 3; //green
this._flasherScaleFactor = 2; // Should this be larger?

//---- Not (yet?) available in config interface ----

this._npcShieldExploder = "N-Shields-explode-star-white";
this._thargoidShieldExploder = "N-Shields-explode-ring-green";

this._shield_charge_energybank_threshold = 0.25; // imitates the corresponding global-settings.plist property

this._debug = true;
this._logging = true;
this._spawnedShipsTimers = []; // will hold timer references for the spawned ships

this._substituteDamagehandlers= []; // Array with the functions NShields assigns to NPC ships as their shipTakingDamage 
                                    // ship event handler, initialized at startUp

this._damageHandlers = []; // Array for references to other NPC shipTakingDamage-like handlers which imitate player
                           // shipTakingDamage, called for all hits but amount=0 if shields still up
                           // For other OXPs to populate

this._wsSC = null; // Shield Cycler worldScript reference
this._wsNG = null; // Naval Grid Next worldScript reference



//==== WorldScript Event Handlers ====//

//-----------------------------------------------------------------------------------------------------
this.startUp = function _NShields_startUp() {
    if (worldScripts["Shield Cycler"]) {
        if (worldScripts["Shield Cycler"]._sc_award_equipment) {
            // Shield Cycler v2.x
            this._wsSC = worldScripts["Shield Cycler"];
        } else
            log(this.name, "Shield Cycler version too old, please upgrade it.");
    }

    if (worldScripts.NavalGrid) {
        if (worldScripts.NavalGrid.$ng_calc_shield_recharge_update)
            // NavalGrid v2.x
            this._wsNG = worldScripts.NavalGrid;
        else
            log(this.name, "Naval Grid Next version too old, please upgrade it.");
    }

    if (worldScripts.NPC_Equipment_Damage) 
        worldScripts.NPC_Equipment_Damage.NPC_equipmentDamagedHandlers.push(this.$NPC_shield_equipmentDamaged);
    if (worldScripts.Lib_Config) worldScripts.Lib_Config._registerSet(this._libSettings);

    // Array with the functions NShields assigns to NPC ships as their shipTakingDamage ship event handler
    this.$substituteDamagehandlers = [ 
        this.$Fake_override_shipTakingDamage, 
        this.$ForeAft_override_shipTakingDamage, 
        this.$DorsalVentral_override_shipTakingDamage 
    ];
}

//-----------------------------------------------------------------------------------------------------
this.startUpComplete = function _NShields_startUpComlete() {
    this.$loadSettings(); // if set using Library OXP config interface
    //TODO: find shield booster stats once not every time needed?
}

//-----------------------------------------------------------------------------------------------------
this.shipSpawned = function _NShields_shipSpawned(ship) {
    if (!ship.script || !this.$shipHasShields(ship) || ship.scanClass == "CLASS_NO_DRAW") return;
    var spawnedShipsTimers = this._spawnedShipsTimers;
    var id = spawnedShipsTimers.length; // array indices start at 0
    var setSpawnedShip = function setSpawnedShipTimed() { this.$setSpawnedShip(id, ship) };
    // give time for other OXPs (EscortDeck!) to change the ship's script before hooking things into it
    // give time for other OXPs (ShipVersion!) to award the Shield Equalizer and/of Shield Capacitors
    this._spawnedShipsTimers[id] = new Timer(this, this.$setSpawnedShip.bind(this, id, ship), 1*Math.random(), 0);
}




//==== Ship Setup ====//


//-----------------------------------------------------------------------------------------------------
this.$setSpawnedShip = function _NShields_setSpawnedShip(id, ship) {
    var spawnedShipsTimers = this._spawnedShipsTimers;
    this.$configureShip(ship);

    delete spawnedShipsTimers[id];
    spawnedShipsTimers[id] = null;

    var i = spawnedShipsTimers.length;
    var count = 0;
    while (i--) if (spawnedShipsTimers[i]) count++;
    if (count === 0) {
        // no active timers, clean the array
        if (this._debug) log(this.name, "No more NPC setup timers running, cleaning the timers array");
        spawnedShipsTimers.length = 0;
    } else if (this._debug) log(this.name, count+" NPC setup timers still running");
}

//-----------------------------------------------------------------------------------------------------
this.$configureShip = function  _NShields_configureShip(ship) { // might be called by other OXPs to set up the ship
    var ws = worldScripts.NShields;
    // TODO: recognise npc_shields = "custom"; script_info?
    if (!ship || !ship.script) return;
    if (!ship.script._NShields_damageHandlers) {
        // haven't being set up yet
        ship.script._NShields_damageHandlers = this._damageHandlers.slice(); // need to explicitly copy, not just reference
    }
    if (ship.script.shipTakingDamage)  {
        var f = ship.script.shipTakingDamage;
        // don't store NShields event handlers if already applied (avoids infinite loop if damage occurs)
        if (ws._substituteDamagehandlers.indexOf(f) === -1)
            if (ship.script._NShields_damageHandlers.indexOf(f) === -1 )
                ship.script._NShields_damageHandlers.push(f); //store existing handler, if any
    }

    if (ship.isFrangible) this.$configureSubEntities(ship);
    
    if (ship.isThargoid) {
        this.$configureThargoidShields(ship);
        return;
    }

    this.$configureNPCShields(ship);
    if (this._debug) log(this.name, ship.displayName+" has "+ship.script._NShields_damageHandlers.length+" damage handlers");
}

//-----------------------------------------------------------------------------------------------------
this.$shipHasShields = function _NShields_shipHasShields(ship) {
    if (ship.scriptInfo && ship.scriptInfo.npc_shields === "no") return false;
    return (ship.script && ship.isPiloted && ["escape-capsule"].indexOf(ship.primaryRole) === -1) // isPiloted rules out buoys, rocks, weapons and debris
}

//-----------------------------------------------------------------------------------------------------
this.$configureNPCShields = function _NShields_configureNPCShields(ship) {
    if (!this._NPC_shields) {
        ship.script.shipTakingDamage = ship.script.$NShields_override_shipTakingDamage = this.$Fake_override_shipTakingDamage;
        return;
    }
    // Standard shield data:
    ship.maxForwardShield = ship.maxAftShield = 128;
    ship.forwardShieldRechargeRate = ship.aftShieldRechargeRate = 2.0;
    
    if (ship.scriptInfo) {
        // Possible non-standard adjustments:
        if (ship.scriptInfo.max_shield_forward) ship.maxForwardShield = parseInt(ship.scriptInfo.max_shield_forward);
        if (ship.scriptInfo.max_shield_aft) ship.maxAftShield = parseInt(ship.scriptInfo.max_shield_aft);
        if (ship.scriptInfo.shield_recharge_forward) ship.forwardShieldRechargeRate = parseFloat(ship.scriptInfo.shield_recharge_forward);
        if (ship.scriptInfo.shield_recharge_aft) ship.aftShieldRechargeRate = parseFloat(ship.scriptInfo.shield_recharge_aft);
        if (ship.scriptInfo.NShields_energyReduction) ship.maxEnergy -= parseInt(ship.scriptInfo.NShields_energyReduction);
    }
    
    ship.script.shipTakingDamage = ship.script.$NShields_override_shipTakingDamage = this.$ForeAft_override_shipTakingDamage;
    ship.script._NShields_lastRechargeTime = 0;
    if (!ship.script._NShields_rechargeTimer)
        ship.script._NShields_rechargeTimer = new Timer(ship.script, worldScripts.NShields.$ForeAft_updateRecharge.bind(ship.script), 1, 1);
    ship.script._NShields_sqRadius = Math.pow((ship.collisionRadius * 1.2), 2);
    
    this.$applyShieldUpgrades(ship);

    if (this._wsNG &&
        (ship.equipmentStatus("EQ_ENERGY_GRID") === "EQUIPMENT_OK" || ship.equipmentStatus("EQ_NAVAL_GRID") === "EQUIPMENT_OK")) {
        // apply Energy Grid upgrades to shield recharge
        var upgraded_shield_recharge = this._wsNG.$ng_calc_shield_recharge_update(
                                                                                    ship, 
                                                                                    {
                                                                                        fwd:ship.forwardShieldRechargeRate, 
                                                                                        aft:ship.aftShieldRechargeRate
                                                                                    }, 
                                                                                    "N-Shield"
                                                                                 );
        ship.forwardShieldRechargeRate = upgraded_shield_recharge.fwd;
        ship.aftShieldRechargeRate = upgraded_shield_recharge.aft;
    }

    ship.script._NShields_flashers = this.$getShieldColours(ship);
    ship.script._NShields_exploder = this.$getShieldExploder(ship);
    ship.script.$NShields_adjustsEnergy = this.$adjustsEnergy;

    ship.forwardShield = ship.maxForwardShield;
    ship.aftShield = ship.maxAftShield;
    if (this._logging)
        log(this.name, ship.displayName+": MaxShields (F/A): "+ship.maxForwardShield.toFixed(0)+"/"+ship.maxAftShield.toFixed(0)+", shiedlRecharge: "+ship.forwardShieldRechargeRate.toFixed(1)+"/"+ship.aftShieldRechargeRate.toFixed(1)+", energy: "+ship.energy.toFixed(1)+", energyRecharge: "+ship.energyRechargeRate.toFixed(1));
}

//-----------------------------------------------------------------------------------------------------
this.$configureThargoidShields = function _NShields_configureThargoidShields(ship) {
    if (!this._Thargoid_shields) {
        ship.script.shipTakingDamage = ship.script.$NShields_override_shipTakingDamage = this.$Fake_override_shipTakingDamage;
        return;
    }
    ship.maxDorsalShield = ship.maxVentralShield = 192;
    ship.dorsalShieldRechargeRate = ship.ventralShieldRechargeRate = 1.5;
    if (this._Thargoid_energyReduction) ship.maxEnergy -= 192;
    if (ship.scriptInfo) {
        if (ship.scriptInfo.max_upper_shield) ship.maxDorsalShield = parseInt(ship.scriptInfo.max_upper_shield);
        if (ship.scriptInfo.max_lower_shield) ship.maxVentralShield = parseInt(ship.scriptInfo.max_lower_shield);
        if (ship.scriptInfo.shield_recharge_upper) ship.dorsalShieldRechargeRate = parseFloat(ship.scriptInfo.shield_recharge_upper);
        if (ship.scriptInfo.shield_recharge_lower) ship.ventralShieldRechargeRate = parseFloat(ship.scriptInfo.shield_recharge_lower);
    }
    ship.script.shipTakingDamage = ship.script.$NShields_override_shipTakingDamage = this.$DorsalVentral_override_shipTakingDamage;
    ship.script._NShields_lastRechargeTime = 0;
    if (!ship.script._NShields_rechargeTimer) 
        ship.script._NShields_rechargeTimer = new Timer(ship.script, worldScripts.NShields.$DorsalVentral_updateRecharge.bind(ship.script), 1, 1);
    ship.script._NShields_sqRadius = Math.pow((ship.collisionRadius * 1.2), 2);
    // No upgrades for Thargoids as the code doesn't support them any more
    ship.script._NShields_flashers = this.$getShieldColours(ship);
    ship.script._NShields_exploder = this.$getShieldExploder(ship);
    ship.dorsalShield = ship.maxDorsalShield;
    ship.ventralShield = ship.maxVentralShield;
}

//-----------------------------------------------------------------------------------------------------
this.$configureSubEntities = function _NShields_configureSubEntities(ship) {
    if (ship.scriptInfo && ship.scriptInfo.NShields_protect_subentities === "no") return;
    //used to allow subEntities to have their own shields, but this is no longer supported
    
    for (sE in ship.subEntities) {
        let subent = ship.subEntities[sE];
        if (subent.script.shipTakingDamage && subent.script.shipTakingDamage != this.$subEntity_override_shipTakingDamage)
            subent.script.$NShields_hold_shipTakingDamage = subent.script.shipTakingDamage;
        subent.script.shipTakingDamage = this.$subEntity_override_shipTakingDamage;
    }
}

//-----------------------------------------------------------------------------------------------------
this.$applyShieldUpgrades = function  _NShields_applyShieldUpgrades(ship) {
    var wsc = this._wsSC;
    // NOTE: may need 'labels' argument for backwards-compatibility?
    if (ship.equipmentStatus("EQ_SHIELD_BOOSTER") === "EQUIPMENT_OK" || 
        ship.equipmentStatus("EQ_NSHIELDS_NPC_SHIELD_BOOSTER") === "EQUIPMENT_OK" ) {
        if (ship.equipmentStatus("EQ_SHIELD_BOOSTER") === "EQUIPMENT_OK") {
            // switches shield booster equipment: the core game ones will increase ship's energy
            ship.removeEquipment("EQ_SHIELD_BOOSTER");
            ship.awardEquipment("EQ_NSHIELDS_NPC_SHIELD_BOOSTER"); 
        }
        this.$shieldBoosterAdjustment(ship, "EQ_SHIELD_BOOSTER");
    }
    if (ship.equipmentStatus("EQ_SHIELD_ENHANCER") === "EQUIPMENT_OK" ||
        ship.equipmentStatus("EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER") === "EQUIPMENT_OK" ||
        ship.equipmentStatus("EQ_NAVAL_SHIELD_BOOSTER") === "EQUIPMENT_OK" ) {
        if (ship.equipmentStatus("EQ_SHIELD_ENHANCER") === "EQUIPMENT_OK") {
            // switches shield booster equipment: the core game ones will increase ship's energy
            ship.removeEquipment("EQ_SHIELD_ENHANCER");
            ship.awardEquipment("EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER");
        }
        if (ship.equipmentStatus("EQ_NAVAL_SHIELD_BOOSTER") === "EQUIPMENT_OK") {
            // switches shield booster equipment: the core game ones will increase ship's energy
            ship.removeEquipment("EQ_NAVAL_SHIELD_BOOSTER");
            ship.awardEquipment("EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER");
        }
        this.$shieldBoosterAdjustment(ship, "EQ_NAVAL_SHIELD_BOOSTER");
    }

    if (wsc && ship.equipmentStatus("EQ_SHIELD_EQUALIZER") !== "EQUIPMENT_OK")
        this.$PossiblyAwardShieldCycler(ship);
 
    if (worldScripts.ShipConfiguration_Core) {
        this.$shipconfAdjustment(ship);
        //worldScripts.ShipConfiguration_Core.$updateShipSpecs(ship);
    }
}

//-----------------------------------------------------------------------------------------------------
this.$shieldBoosterAdjustment = function  _NShields_shieldBoosterAdjustment(ship, type, reverse) {
    //var shield_boost = this.shieldBoosters[type].shield_boost;
    //var recharge_boost = this.shieldBoosters[type].recharge_boost;
    let eq = EquipmentInfo.infoForKey(type);
    let shield_boost = parseInt(eq.scriptInfo.oolite_shield_increase);
    let recharge_boost = parseFloat(eq.scriptInfo.oolite_shield_recharge_multiplier);
    if (!recharge_boost) recharge_boost = 1; //just in case, to avoid NaN and divide-by-zero errors
    if (ship.script.$debug)
        log(this.name, ship.displayName+" - Shields before:"+ship.maxForwardShield+"/"+ship.maxAftShield+", eq:"+type+", boost:"+shield_boost+", recharge_boost:"+recharge_boost);
    if (reverse) {
        // damaging/removing shield booster, not awarding
        ship.maxForwardShield -= shield_boost;
        ship.maxAftShield -= shield_boost;
        ship.forwardShieldRechargeRate /= recharge_boost;
        ship.aftShieldRechargeRate /= recharge_boost;
        if (ship.forwardShield > ship.maxForwardShield) ship.forwardShield = ship.maxForwardShield;
        if (ship.aftShield > ship.maxAftShield) ship.aftShield = ship.maxAftShield;
    } else {
        ship.maxForwardShield += shield_boost;
        ship.maxAftShield += shield_boost;
        ship.forwardShieldRechargeRate *= recharge_boost;
        ship.aftShieldRechargeRate *= recharge_boost;
        ship.forwardShield = ship.maxForwardShield; // in case this was called after another script awarded equipment
        ship.aftShield = ship.maxAftShield;
    }
    if (ship.script.$debug)
        log(this.name, ship.displayName+" - Shields after:"+ship.maxForwardShield+"/"+ship.maxAftShield+", recharge:"+ship.forwardShieldRechargeRate+"/"+ship.aftShieldRechargeRate);
}

//-----------------------------------------------------------------------------------------------------
this.$getShieldColours = function _NShields_getShieldColours(ship) {
    if (ship.scriptInfo) {
        if (ship.scriptInfo.NShields_flasherKeys) return ship.scriptInfo.NShields_flasherKeys;
        //TODO: name not number for script_info?
        if (ship.scriptInfo.NShields_flasherPreset) return this._flasherColours[ship.scriptInfo.NShields_flasherPreset];    
    }
    if (ship.isThargoid) return this._flasherColours[this._Thargoid_flasherColours];
    return this._flasherColours[this._NPC_flasherColours];
}

//-----------------------------------------------------------------------------------------------------
this.$getShieldExploder = function _NShields_getShieldExploder(ship) {
    if (ship.collisionRadius > 128) var sizeClass = "-XL";
    else if (ship.collisionRadius > 64) var sizeClass = "-L";
    else if (ship.collisionRadius > 32) var sizeClass = "-M";
    else if (ship.collisionRadius > 16) var sizeClass = "-S";
    else var sizeClass = "-XS";
    
    if (ship.scriptInfo && ship.scriptInfo.NShields_exploders)
    {
        return ship.scriptInfo.NShields_exploders[0]+sizeClass;
    }
    if (ship.isThargoid) return this._thargoidShieldExploder+sizeClass;
    return this._npcShieldExploder+sizeClass;
}



//==== Shield Simulation ====//
// For these functions 'this' must mean a ship script 

//---- Standard fore/aft shields: ----

//-----------------------------------------------------------------------------------------------------
this.$ForeAft_override_shipTakingDamage = function _NShields_ForeAft_override_shipTakingDamage(amount, whom, type, caller) { 
    //'caller' is added when protecting frangible subEntities 
    var that = _NShields_ForeAft_override_shipTakingDamage;
    var wns = (that.wns = that.wns || worldScripts.NShields);
    var adjustsEnergy = (that.adjustsEnergy = that.adjustsEnergy || wns.$adjustsEnergy);
    var ship = this.ship; // th NPC ship
    var hitLocation, hitLocationRelative, hitLocationString, energy_before;
    var logging = (this.$logging || this.ship == player.ship.target);

    function applyDamageToShield(hit_location, hit_location_string, damage, logging) {
        var damage_subent;
        var shield_name = (hit_location_string === "Aft" ? "aftShield" : "forwardShield");
        ship[shield_name] -= damage;
        if (ship[shield_name] <= 0) {
            if (caller) {
                // take the damage from the subentity, as much as it can take, keep the balance as deficit in shield
                // remember: forwardShield is negative! (in defict)
                var damage_subent = (caller.energy >= -ship[shield_name] ? -ship[shield_name] : caller.energy);
                if (caller.script.$NShields_hold_shipTakingDamage)
                    caller.script.$NShields_hold_shipTakingDamage.bind(caller.script)(damage_subent, whom, type);
                else {
                    energy_before = caller.energy;
                    adjustsEnergy(caller, -damage_subent, "(ForeAft_override_shipTakingDamage apply damage)");  
                    if (logging) log("NShields", "taking "+damage_subent.toFixed(1)+" from "+caller.displayName+" energy, from "+energy_before.toFixed(1)+" to "+caller.energy.toFixed(1));
                }
                ship[shield_name] += damage_subent;
            }
            if (logging) log("NShields", ship.displayName+": "+hit_location_string+"Sh zeroed, energy "+ship.energy.toFixed(1));
            // propagate the non-absorbed damage downstream
            let remaining_damage = -ship[shield_name];
            ship[shield_name] = 0;
            wns.$applyDamageHandlers.call(ship.script, ship, remaining_damage, whom, type)
            if (ship.isVisible) wns.$drawSparks(hit_location, ship, remaining_damage);
        } else 
            wns.$applyDamageHandlers.call(ship.script, ship, 0, whom, type);
        if (ship.isVisible) wns.$drawShieldFX(hit_location, ship[shield_name], ship, whom)
        return;
    }

    if (amount == 0 || type == "hit a planet" || type == "hit a sun" || type == "cascade weapon" || type == "removed") {
        //cases where, for practical purposes, no shield handling is needed - but other OXPs' handlers may still need to run
        wns.$applyDamageHandlers(ship, amount, whom, type)
        return;
    }

    if (logging) log("NShields", ship.displayName+" has "+ship.energy.toFixed(1)+" energy and shields "+ship.forwardShield.toFixed(1)+"/"+ship.aftShield.toFixed(1)+", damage: "+amount.toFixed(1));
   
    // recharge shields for the time interval since last recharge 
//    let recharge_interval = (!this.NShields_lastRechargeTime ? 0 : clock.absoluteSeconds - this.NShields_lastRechargeTime);
//    worldScripts.NShields.ForeAft_updateRecharge.apply(this, [recharge_interval]);
//    worldScripts.NShields.$resetRechargeTimer.call(this);
    energy_before = ship.energy;
    if (caller) 
        //heal subEntity if called from there
        adjustsEnergy(caller, amount, "(ForeAft_override_shipTakingDamage heal back)");  
    else 
        adjustsEnergy(ship, amount, "(ForeAft_override_shipTakingDamage heal back)");
    if (logging) log("NShields", ship.displayName+" now has "+ship.energy.toFixed(1)+" energy and shields "+ship.forwardShield.toFixed(1)+"/"+ship.aftShield.toFixed(1));
    
    if (type === "heat damage") {// will not be able to locate hit, and different handling is required anyway
        wns.$applyHeatDamage(ship, amount, ["forwardShield", "aftShield"]);
        return;
    }
    
    hitLocation = wns.$locateHit(ship, whom)
    hitLocationRelative = hitLocation.subtract(ship.position);
    hitLocationString = hitLocationRelative.dot(ship.vectorForward) > 0 ? "Fwd" : "Aft";

    if (logging) log("NShields", ship.displayName+": "+(ship.energy?"":"ZERO_ENERGY ")+"taking "+amount.toFixed(1)+" damage of type "+type+" from "+(whom ? whom.displayName : "Unknown")+", "+hitLocationString+"Sh from "+(hitLocationString === "Fwd" ? ship.forwardShield.toFixed(0)+" to "+(ship.forwardShield-amount).toFixed(0) : ship.aftShield.toFixed(0)+" to "+(ship.aftShield-amount).toFixed(0))+", energy "+ship.energy.toFixed(1)+"("+energy_before.toFixed(1)+")"+(caller ? " at "+caller.displayName :""));

    applyDamageToShield(hitLocation, hitLocationString, amount, logging);

    if (logging) log("NShields", ship.displayName+" after damage handlers: shields "+ship.forwardShield.toFixed(1)+"/"+ship.aftShield.toFixed(1)+", energy: "+ship.energy.toFixed(1));

    if (ship.energy <= 0 && ship.script._NShields_rechargeTimer && ship.script._NShields_rechargeTimer.isRunning) {
        ship.script._NShields_rechargeTimer.stop();
        delete ship.script._NShields_rechargeTimer;
    }
}

//-----------------------------------------------------------------------------------------------------
this.$ForeAft_updateRecharge = function _NShields_ForeAft_updateRecharge() {
    var that = _NShields_ForeAft_updateRecharge;
    var wNS = (that.wNS = that.wNS || worldScripts.NShields);
    var shield_charge_energybank_threshold = (that.shield_charge_energybank_threshold = that.shield_charge_energybank_threshold ||
                                                wNS._shield_charge_energybank_threshold);
    var pship = (that.pship = that.pship || player.ship);
    var ship = this.ship;
    var now = clock.absoluteSeconds;
    var last = this._NShields_lastRechargeTime;
    var delta = now - last;
    this._NShields_lastRechargeTime = now;

    if (ship.energy < ship.maxEnergy * shield_charge_energybank_threshold) return;

    var energyForShields = ship.energy;
    var fwdMax = ship.maxForwardShield;
    var aftMax = ship.maxAftShield;
    var shieldRechargeFwd = ship.forwardShieldRechargeRate * delta;
    var shieldRechargeAft = ship.aftShieldRechargeRate * delta;
    
    var rechargeFwd = Math.min(shieldRechargeFwd, fwdMax - ship.forwardShield);
    var rechargeAft = Math.min(shieldRechargeAft, aftMax - ship.aftShield);
    
    var old_fwd = ship.forwardShield, old_aft = ship.aftShield;
    if ((ship.forwardShield > fwdMax * 0.25) && (ship.aftShield > aftMax * 0.25))
        energyForShields = Math.max(0.0, ship.energy - 0.1 - (ship.maxEnergy * shield_charge_energybank_threshold));
    
    if (ship.forwardShield < ship.aftShield) {
        rechargeFwd = Math.min(rechargeFwd, energyForShields);
        rechargeAft = Math.min(rechargeAft, energyForShields - rechargeFwd);
    } else {
        rechargeAft = Math.min(rechargeAft, energyForShields);
        rechargeFwd = Math.min(rechargeFwd, energyForShields - rechargeAft);
    }
    
    ship.forwardShield += rechargeFwd;
    ship.aftShield += rechargeAft;
    var energy_before = ship.energy;
    if (isNaN(rechargeFwd+rechargeAft))
        // prevents "invulnerable to lasers" bug
        log("NShields", ship.displayName+": shield recharge failed, invalid recharge value: "+rechargeFwd+"/"+rechargeAft);
    else
        ship.energy -= (rechargeFwd + rechargeAft);

    if (ship.energy == 0) log("NShields", ship.displayName+" - ZERO_ENERGY shield recharge: recharge F/A:"+rechargeFwd.toFixed(1)+"/"+rechargeAft.toFixed(1)+", energy "+energy_before.toFixed(1)+"->"+ship.energy.toFixed(1)+", energyForShields:"+energyForShields.toFixed(1)+", shields F/A:"+old_fwd.toFixed(1)+"/"+old_aft.toFixed(1)+"->"+ship.forwardShield.toFixed(1)+"/"+ship.aftShield.toFixed(1)+", origShRecharge (F/A):"+(ship.forwardShieldRechargeRate * delta).toFixed(1)+"/"+(ship.aftShieldRechargeRate * delta).toFixed(1)+", delta:"+delta.toFixed(5));
    // Oolite's source code does something called "OOClamp_0_max_f" on the shield and energy values here. I have not included any analogous code because I don't know what the aforementioned function actually does. Hopefully this will not break things.
    if (ship.forwardShield >= fwdMax && ship.aftShield >= aftMax) this._NShields_rechargeTimer.stop();
}


//-----------------------------------------------------------------------------------------------------
this.$DorsalVentral_override_shipTakingDamage = function _NShields_DorsalVentral_override_shipTakingDamage(amount, whom, type, caller) {
    // Thargoid dorsal/ventral shields: ----
    var that = _NShields_DorsalVentral_override_shipTakingDamage;
    var wsNSh = (that.wsNSh = that.wsNSh || worldScripts.NShields);
    var adjustsEnergy = (that.adjustsEnergy = that.adjustsEnergy || wsNSh.$adjustsEnergy);
    var ship = this.ship;

    if (amount == 0 || type == "hit a planet" || type == "hit a sun" || type == "cascade weapon" || type == "removed") {
        wsNSh.$applyDamageHandlers(ship, amount, whom, type)
        return;
    }
    
//    wsNSh.$DorsalVentral_updateRecharge.apply(this, [clock.absoluteSeconds - this._NShields_lastRechargeTime]);
//    wsNSh.$resetRechargeTimer.call(this);
    if (caller)
        //heal subEntity if called from there
        adjustsEnergy(caller, amount, "(DorsalVentral_override_shipTakingDamage heal back)");
    else 
        adjustsEnergy(ship, amount, "(DorsalVentral_override_shipTakingDamage heal back)");
    
    if (type == "heat damage") {
        ws.NSh.$applyHeatDamage(ship, amount, ["dorsalShield", "ventralShield"]);
        return;
    }
    
    var hitLocation = wsNSh.$locateHit(ship, whom)
    var hitLocationRelative = hitLocation.subtract(ship.position);
    if (hitLocationRelative.dot(ship.vectorUp) > 0) {
        // dorsal shield
        ship.dorsalShield -= amount;
        if (ship.dorsalShield < 0) {
            if (caller) 
                adjustsEnergy(caller, ship.dorsalShield, "(DorsalVentral_override_shipTakingDamage apply damage)");
            else 
                // add negative shield value to energy
                adjustsEnergy(ship, ship.dorsalShield, "(DorsalVentral_override_shipTakingDamage apply damage)");
            wsNSh.$applyDamageHandlers(ship, -ship.dorsalShield, whom, type)
            if (ship.isVisible) wsNSh.$drawSparks(hitLocation, ship, -ship.dorsalShield);
            ship.dorsalShield = 0;
        }
        else wsNSh.$applyDamageHandlers(ship, 0, whom, type);
        if (ship.isVisible) wsNSh.$drawShieldFX(hitLocation, ship.dorsalShield, ship, whom)
    } else {
        ship.ventralShield -= amount;
        if (ship.ventralShield < 0) {
            if (caller) 
                adjustsEnergy(caller, ship.ventrallShield, "(DorsalVentral_override_shipTakingDamage apply damage)");
            else 
                adjustsEnergy(ship, ship.ventralShield, "(DorsalVentral_override_shipTakingDamage apply damage)");
            wsNSh.$applyDamageHandlers(ship, -ship.ventralShield, whom, type)
            if (ship.isVisible) wsNSh.$drawSparks(hitLocation, ship, -ship.ventralShield);
            ship.ventralShield = 0;
        }
        else wsNSh.$applyDamageHandlers(ship, 0, whom, type);
        if (ship.isVisible) wsNSh.$drawShieldFX(hitLocation, ship.ventralShield, ship, whom)
    }    
    if (ship.energy <= 0 && ship.script._NShields_rechargeTimer && ship.script._NShields_rechargeTimer.isRunning) {
            ship.script._NShields_rechargeTimer.stop();
            delete ship.script._NShields_rechargeTimer;
    }
}

//-----------------------------------------------------------------------------------------------------
this.$DorsalVentral_updateRecharge = function _NShields_DorsalVentral_updateRecharge(delta) {
    //This one is a bit simpler as it doesn't have to precisely mimic player shields
    var ship = this.ship;

    if (delta == undefined) delta = 1;
    this._NShields_lastRechargeTime = clock.absoluteSeconds;
    var topMax = ship.maxDorsalShield;
    var btmMax = ship.maxVentralShield;
    var rechargeTop = Math.min(ship.dorsalShieldRechargeRate * delta, topMax - ship.dorsalShield);
    var rechargeBtm = Math.min(ship.ventralShieldRechargeRate * delta, btmMax - ship.ventralShield);
    if (ship.dorsalShield < ship.ventralShield) {
        rechargeTop = Math.min(rechargeTop, ship.energy);
        rechargeBtm = Math.min(rechargeBtm, ship.energy - rechargeTop);
    } else {
        rechargeBtm = Math.min(rechargeBtm, ship.energy);
        rechargeTop = Math.min(rechargeTop, ship.energy - rechargeBtm);
    }
    ship.dorsalShield += rechargeTop;
    ship.ventralShield -= rechargeBtm;
    if (isNaN(rechargeTop+rechargeBtm))
        // prevents "invulnerable to lasers" bug
        log("NShields", ship.displayName+": shield recharge failed, invalid recharge value: "+rechargeTop+"/"+rechargeBtm);
    else
        ship.energy -= (rechargeTop + rechargeBtm);
    if (ship.dorsalShield == topMax && ship.ventralShield == btmMax) this._NShields_rechargeTimer.stop();
}


//---- Sub-entity shielding: ----

//-----------------------------------------------------------------------------------------------------
this.$subEntity_override_shipTakingDamage = function _NShields_subEntity_override_shipTakingDamage(amount, whom, type) {
    var owner = this.ship.owner;
    if (owner.script.$logging || owner == player.ship.target)
        log("NShields", owner.displayName+" - subEntity "+this.ship.displayName+" taking "+amount+" damage of type "+type+" from "+(whom ? whom.displayName : "Uknown"));
    owner.script.$NShields_override_shipTakingDamage(amount, whom, type, this.ship) //not simple 'shipTakingDamage' in case another OXP re-overrode the N-Shields override
}


//---- Fake shield effects based on energy level ----

//-----------------------------------------------------------------------------------------------------
this.$Fake_override_shipTakingDamage = function _NShields_Fake_override_shipTakingDamage(amount, whom, type, caller) {
    var ship = this.ship;
    var wsNSh = worldScripts.NShields;

    var fakeShield = ship.energy - Math.min(64, ship.energy / 2);
    wsNSh.$applyDamageHandlers(ship, amount, whom, type)
    if (ship.isVisible) {
        var hitLocation = wsNSh.$locateHit(ship, whom);
        if (fakeShield > 0) wsNSh.$drawShieldFX(hitLocation, fakeShield, ship, whom);
        else wsNSh.$drawSparks(hitLocation, ship, amount);
    }
}


//---- Common damage handler components: ----

//-----------------------------------------------------------------------------------------------------
this.$adjustsEnergy = function _NShields_adjustsEnergy(ship, amount, caller) {
    if (isNaN(amount))
        log("NShields", ship.displayName+": energy adjustement from "+caller+" failed, invalid value: "+amount+", energy stays at "+ship.energy.toFixed(1));
    else
        ship.energy += amount
}

//-----------------------------------------------------------------------------------------------------
this.$locateHit = function _NShields_locateHit(ship, attacker) {
    var myscript = ship.script;
    var sqRadius = myscript._NShields_sqRadius;
    //loosely based on customshields
    if (!attacker) {
        var cloaked_attacker = myscript._NShields_cloaked_attacker;
        if (player.ship.isCloaked && ship.isVisible) attacker = player.ship;
        else if (cloaked_attacker && cloaked_attacker.position && cloaked_attacker.position.distanceTo(ship) < 25600) attacker = cloaked_attacker;
        else attacker = myscript._NShields_cloaked_attacker = this.$findCloakedAttacker(ship);
        if (!attacker) {
            //still not found anyone?
//            log(this.name, ship.displayName+": WARNING: attacker is null; could not find cloaked attacker")
            return ship.position;
        }
    }
    var range = attacker.position.distanceTo(ship) - (ship.collisionRadius + 1);
    var hitLocation = attacker.position.add(attacker.vectorForward.multiply(range)); // possible forward-weapon hit
    if (ship.position.squaredDistanceTo(hitLocation) > sqRadius) {
        // More efficient to find & compare squares as distanceTo() would involve a square root
        hitLocation = attacker.position.add(attacker.vectorForward.multiply(-range)); // next try aft weapon
        if (ship.position.squaredDistanceTo(hitLocation) > sqRadius) {
            hitLocation = attacker.position.add(attacker.vectorRight.multiply(range));
            if (ship.position.squaredDistanceTo(hitLocation) > sqRadius) {
                hitLocation = attacker.position.add(attacker.vectorRight.multiply(-range));
                if (ship.position.squaredDistanceTo(hitLocation) > sqRadius) {
                    hitLocation = Vector3D.interpolate(ship.position, attacker.position, ship.collisionRadius / attacker.position.distanceTo(ship)); // fallback method, directly between ships
                }
            }
        }
    }
    return hitLocation;
}

//-----------------------------------------------------------------------------------------------------
this.$findCloakedAttacker = function _NShields_findCloakedAttacker(ship) {
    function $isCloakedAttacker(entity) {
        return (entity.isShip && !entity.isPlayer && entity.isCloaked && entity.target);
    }
    var scan = system.filteredEntities(this, $isCloakedAttacker, ship, 25600);
    for (i in scan) {
        if (scan[i].target == ship) return scan[i];
    }
}

//-----------------------------------------------------------------------------------------------------
this.$resetRechargeTimer = function _NShields_resetRechargeTimer() {
    // called with 'this' pointing to ship script
    this._NShields_rechargeTimer.stop()
    this._NShields_rechargeTimer.nextTime = clock.absoluteSeconds + 1;
    this._NShields_rechargeTimer.start()
}

//-----------------------------------------------------------------------------------------------------
this.$applyHeatDamage = function _NShields_applyHeatDamage(ship, amount, shields) {
    var pass = 0; //amount passed on to energy banks
    
    for (S in shields) {
        var shield_amt = amount / shields.length;
        ship[shields[S]] -= shield_amt;
        if (ship[shields[S]] < 0) {
            pass += -ship[shields[S]];
            ship[shields[S]] = 0;
        }
    }
//    ship.energy -= pass;
    this.$applyDamageHandlers.call(this, ship, pass, null, "heat damage")
}

//-----------------------------------------------------------------------------------------------------
this.$applyDamageHandlers = function _NShields_applyDamageHandlers(ship, amount, whom, type) {
    var that = _NShields_applyDamageHandlers;
    var wsNSh = (that.wsNSh = that.wsNSh || worldScripts.NShields);
    var adjustsEnergy = (that.adjustsEnergy = that.adjustsEnergy || wsNSh.$adjustsEnergy);
    var logging = (this.$logging || ship == player.ship.target);
    var damageHandlers = ship.script._NShields_damageHandlers;
    var i = damageHandlers.length;

    if (amount)
        if (i === 0 || (i === 1 && damageHandlers[0] === wsNSh.$NPC_sc_adjust)) {
            var old_energy = ship.energy;
            adjustsEnergy(ship, -amount, "($applyDamageHandlers apply damage)");
            if (logging) log("NShields",  ship.displayName+" taking "+amount+" from ship's energy: "+old_energy.toFixed(1)+"->"+ship.energy.toFixed(1));
            amount = 0;
        }
    while (i--) {
        if (logging || ship == player.ship.target) log("NShields", ship.displayName+", calling damage handler "+damageHandlers[i].name+" with "+amount.toFixed(1)+" damage of type "+type+", energy: "+ship.energy.toFixed(1));
        damageHandlers[i].apply(ship.script, [amount, whom, type]);
        if (logging) log("NShields", ship.displayName+" after damage handler: shields "+ship.forwardShield.toFixed(1)+"/"+ship.aftShield.toFixed(1)+", energy: "+ship.energy.toFixed(1));
    }
}


//==== Visual Effects ===================================================================

//-----------------------------------------------------------------------------------------------------
this.$drawShieldFX = function _NShields_drawShieldFX(position, shieldLevel, ship, attacker) {
    var flashers = ship.script._NShields_flashers;
    if (shieldLevel < 1 || !flashers) return; // shields effectively down, no draw
    var colour_idx = Math.min(Math.floor(shieldLevel/32), (flashers.length - 1));
    if (this._drawFlashers) {
        var effectType = "N-Shields_flasher_" + flashers[colour_idx];
        var shieldFlasher = system.addVisualEffect(effectType,  position);
        if (attacker) shieldFlasher.orientation = new Vector3D().rotationTo(attacker.position.subtract(ship.position));
        shieldFlasher.script.$parent = ship;
        shieldFlasher.scale(ship.collisionRadius * this._flasherScaleFactor); //TODO: copy more sophisticated scale code?
    }
    if (this._drawExploders) {
        var expl = ship.spawnOne(ship.script._NShields_exploder);
        if (expl) {
            expl.position = position;
            expl.explode();
        }
    }
}

//-----------------------------------------------------------------------------------------------------
this.$drawSparks = function _NShields_drawSparks(position, ship, amount) {
    if (!this._drawSparkEffects) return;
    
    var chance = 1;
    if (amount < 10) chance = 0.5;
    if (Math.random() <= chance) {
        var sparks = ship.spawnOne("NShields-sparks");
        sparks.position = position;
        sparks.explode()
    }
}

//---- Flasher lists: ----
this._multicolour1 = [
    "red",
    "orange",
    "yellow",
    "green",
    "cyan",
    "cyan",
    "blue",
    "blue",
    "purple"
];
this._multicolour2 = [
    "red2",
    "yellow2",
    "yellowgreen2",
    "green2",
    "turquoise2",
    "turquoise2",
    "blue2",
    "blue2",
    "purple2"
];
this._multicolour3 = [ //inverted: red high, blue low
    "blue3",
    "turquoise3",
    "green3",
    "yellow3",
    "yellow3",
    "orange3",
    "red3",
    "red3",
    "deep-red"
];
// next one is aimed at Thargoids but can be used on other ships
this._green = [
    "Green01",
    "Green02",
    "Green03",
    "Green04",
    "Green05",
    "green2" // re-use of green flasher from _multicolour2
];
this._blue = [
    "Blue01",
    "Blue02",
    "Blue03",
    "Blue04",
    "Blue05",
    "blue2",
    "Blue07"
];
// TODO: pink, white?
this._flasherColours = [this._multicolour1, this._multicolour2, this._multicolour3, this._green, this._blue];




//==== Compatibility with other OXPs ====================================================

//---- Handler for NPC Equipment Damage ----

//-----------------------------------------------------------------------------------------------------
this.$NPC_shield_equipmentDamaged = function _NShields_NPC_shield_equipmentDamaged(equipment) {
    var shipconf = worldScripts.ShipConfiguration_Core;
    var wsNSh =  worldScripts.NShields;
    var ship = this.ship;

    if (shipconf) {
        if (shipconf._frontshields.indexOf(equipment) >= 0 || shipconf._aftshields.indexOf(equipment) >= 0) {
            // Damage shield boosters etc. first if available
            if (ship.equipmentStatus("EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER") === "EQUIPMENT_OK") {
                ship.setEquipmentStatus(equipment, "EQUIPMENT_OK")
                ship.setEquipmentStatus("EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER", "EQUIPMENT_DAMAGED")
                equipment = "EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER"; // so later code will immediately deactivate the damaged booster
            } else if (ship.equipmentStatus("EQ_NSHIELDS_NPC_SHIELD_BOOSTER") === "EQUIPMENT_OK") {
                ship.setEquipmentStatus(equipment, "EQUIPMENT_OK")
                ship.setEquipmentStatus("EQ_NSHIELDS_NPC_SHIELD_BOOSTER", "EQUIPMENT_DAMAGED")
                equipment = "EQ_NSHIELDS_NPC_SHIELD_BOOSTER";
            } else {
                // actually damage the basic shield generator
                if (shipconf._frontshields.indexOf(equipment) >= 0) ship.forwardShieldRechargeRate = -0.5;
                if (shipconf._aftshields.indexOf(equipment) >= 0) ship.aftShieldRechargeRate = -0.5;
            }
        }
    }
    
    if (equipment === "EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER") //reverse the equipment upgrade changes
        wsNSh.$shieldBoosterAdjustment(this.ship, "EQ_NAVAL_SHIELD_BOOSTER", true);
    
    if (equipment === "EQ_NSHIELDS_NPC_SHIELD_BOOSTER")
        wsNSh.$shieldBoosterAdjustment(this.ship, "EQ_SHIELD_BOOSTER", true);
}

//---- Ship Configuration shield equipment ----

//-----------------------------------------------------------------------------------------------------
this.$shipconfAdjustment = function _NShields_shipconfAdjustment(ship) {
    var wsSConf = worldScripts.ShipConfiguration_Core;
    var shipconf_fwdShield = wsSConf.$equipmentItemInUse("frontshields", ship);
    var shipconf_aftShield = wsSConf.$equipmentItemInUse("aftshields", ship);
    
    if (shipconf_fwdShield && shipconf_fwdShield != ship.script.NShields_shipconf_fwdShield) //in case we ran this function already
    {
        if (shipconf_fwdShield === "EQ_NO_FORWARD_SHIELD") ship.maxForwardShield = 0;
        else ship.maxForwardShield += parseInt(EquipmentInfo.infoForKey(shipconf_fwdShield).scriptInfo.oolite_shield_increase);
        ship.script.NShields_shipconf_fwdShield = shipconf_fwdShield;
        ship.forwardShield = ship.maxForwardShield; // Note this only works reliably as long as we only change equip at setup not in-flight
    }
    if (shipconf_aftShield && shipconf_aftShield != ship.script.NShields_shipconf_aftShield)
    {
        if (shipconf_aftShield === "EQ_NO_AFT_SHIELD") ship.maxAftShield = 0;
        else ship.maxAftShield += parseInt(EquipmentInfo.infoForKey(shipconf_aftShield).scriptInfo.oolite_shield_increase);
        ship.script.NShields_shipconf_aftShield = shipconf_aftShield;
        ship.aftShield = ship.maxAftShield;
    }
}


//---- Shield Cycler ----

//-----------------------------------------------------------------------------------------------------
this.$chooseShieldCycler = function _NShields_chooseShieldCycler(ship) {
    var sc_version, sc_eqKey, manual_version, manual_eqKey;
    if (ship.equipmentStatus("EQ_NSHIELDS_NPC_NAVAL_SHIELD_BOOSTER") === "EQUIPMENT_OK") {
        sc_eqKey = "EQ_SC_SHIELD_CYCLER_ADVANCED";
        sc_version = 3;
        manual_eqKey = "EQ_SC_MANUAL_CONFIGURATOR_STANDARD";
        manual_version = 2;
    } else if (ship.equipmentStatus("EQ_NSHIELDS_NPC_SHIELD_BOOSTER") === "EQUIPMENT_OK") {
        sc_eqKey = "EQ_SC_SHIELD_CYCLER_STANDARD";
        sc_version = 2;
        manual_eqKey = "EQ_SC_MANUAL_CONFIGURATOR_BASIC";
        manual_version = 1;
    } else {
        sc_eqKey: "EQ_SC_SHIELD_CYCLER_BASIC";
        sc_version = 1;
        manual_version = (Math.random() < 0.6) ? 1 : 0;
        if (manual_version) manual_eqKey = "EQ_SC_MANUAL_CONFIGURATOR_BASIC";
    }
    return { 
                sc_eqKey: sc_eqKey,
                version: sc_version,
                manual_eqKey: manual_eqKey, 
                manual_version: manual_version,
            };
}

//-----------------------------------------------------------------------------------------------------
this.$PossiblyAwardShieldCycler = function _NShields_PossiblyAwardShieldCycler(ship) {
    var wsc = this._wsSC;
    var sc_award = (wsc ? wsc._sc_award_equipment : null);
    var hasSC = false;
    var sc;
    if (!sc_award) return;

    // Shield Cycler v2.x installed
    if (!ship.script.$sc_settings) {
        // ship doesn't have Shield Cycler, possibly award it
        if (ship.autoWeapons && Math.random() <= 0.1) {
            sc = this.$chooseShieldCycler(ship);
            if (sc.sc_eqKey) {
                if (this._logging) log(this.name, ship.displayName+" awarding Shield Cycler "+sc.sc_eqKey+(sc.manual_eqKey ? " and "+sc.manual_eqKey : ""));
                sc_award.call(ship.script, sc.sc_eqKey);
                hasSC = true;
                if (sc.manual_eqKey) {
                    sc_award.call(ship.script, sc.manual_eqKey);
                }
            }
        }
    } else
        hasSC = true;

    if (hasSC) {
        if (this._debug && sc) log(this.name, ship.displayName+", version awarded: "+sc.version+", manual configurator version: "+sc.manual_version)
        if (this._logging) log(this.name, ship.displayName+", Enabling ShieldCycler for "+ship.displayName);
        if (ship.script._NShields_damageHandlers.indexOf(this.$NPC_sc_adjust) === -1) 
            ship.script._NShields_damageHandlers.push(this.$NPC_sc_adjust)  
    } else 
        if (this._logging) log(this.name, ship.displayName+" hasn't Shield Cycler");
}

//-----------------------------------------------------------------------------------------------------
this.$NPC_sc_adjust = function _NShields_NPC_sc_adjust(amount, whom, type) {
    // shipTakingDamage handler component for NPC Shield Cyclers
    // 'this' must be the ship script
    var that = _NShields_NPC_sc_adjust;
    var wsNSh = (that.wsNSh = that.wsNSh || worldScripts.NShields);
    var wsSC = (that.wsSC = that.wsSC || wsNSh._wsSC);
    var adjustsEnergy = (that.adjustsEnergy = that.adjustsEnergy || wsNSh.$adjustsEnergy);
    var ship = this.ship;
    var logging = (this.$logging || ship == player.ship.target);

    if (wsSC) {
        // Shield Cycler v2.x only
        wsSC._sc_sc_adjust.call(ship.script, false,  "N-Shields");
        if (logging) log("NShields", ship.displayName+" after SC shields "+ship.forwardShield.toFixed(1)+"/"+ship.aftShield.toFixed(1)+", energy: "+ship.energy.toFixed(1));
    }
}



//-----------------------------------------------------------------------------------------------------
//---- Library OXP config interface ----

this._libSettings = {
    Name: this.name,
    Display: "N-Shields",
    Alive: "_libSettings",
    Notify: "$saveSettings",
    Bool: {
        B0: {
            Name: "_NPC_shields",
            Def: true,
            Desc: "NPC fore/aft shield simulation on/off."
        },
        B1: {
            Name: "_Thargoid_shields",
            Def: true,
            Desc: "Thargoid dorsal/ventral shields on/off."
        },
        B2: {
            Name: "_Thargoid_energy_reduction",
            Def: true,
            Desc: "Reduce Thargoid energy banks to balance their shields."
        },
        B3: {
            Name: "_drawFlashers",
            Def: true,
            Desc: "flasher effects on/off."
        },
        B4: {
            Name: "_drawExploders",
            Def: false,
            Desc: "star-explosion effects on/off."
        },
        B5: {
            Name: "_drawSparkEffects",
            Def: true,
            Desc: "unshielded hit spark-shower effects on/off."
        }
    },
    SInt: {
        S0: {
            Name: "_NPC_flasherColours",
            Def: 1,
            Min: 0,
            Max: this._flasherColours.length,
            Desc: "Colour option for standard NPC (and player) shield flasher effects.",
            Float: false
        },
        S1: {
            Name: "_Thargoid_flasherColours",
            Def: 3,
            Min: 0,
            Max: this._flasherColours.length,
            Desc: "Colour option for Thargoid shield flasher effects.",
            Float: false
        },
        S2: {
            Name: "_flasherScaleFactor",
            Def: 2,
            Min: 0.2,
            Max: 8,
            Desc: "Scale factor for shield flasher effects.",
            Float: true
        },
        Info: "Colours: 0: Multicolour, 1: Multicolour (type 2), 2: Reverse Multicolour (red high, blue low), 3: Thargoid Green, 4: Blue."
    }
}

//-----------------------------------------------------------------------------------------------------
this.$saveSettings = function _NShields_saveSettings() {
    var std = true;
    for (i in this._libSettings.Bool) {
        if (this[this._libSettings.Bool[i].Name] != this._libSettings.Bool[i].Def) std = false;
    }
    for (i in this._libSettings.SInt) {
        if (this[this._libSettings.SInt[i].Name] != this._libSettings.SInt[i].Def) std = false;
    }
    if (!std) {
        var settings = {};
        for (i in this._libSettings.Bool) {
            let name = this._libSettings.Bool[i].Name;
            settings[name] = this[name];
        }
        for (i in this._libSettings.SInt) {
            let name = this._libSettings.SInt[i].Name;
            settings[name] = this[name];
        }
        missionVariables.NShields_settings = JSON.stringify(settings);
    }
    else missionVariables.NShields_settings = null;
}

//-----------------------------------------------------------------------------------------------------
this.$loadSettings = function _NShields_loadSettings() {
    var save = JSON.parse(missionVariables.NShields_settings);
    if (save) {
        for (i in this._libSettings.Bool) {
            let name = this._libSettings.Bool[i].Name;
            this[name] = save[name];
        }
        for (i in this._libSettings.SInt) {
            let name = this._libSettings.SInt[i].Name;
            this[name] = save[name];
        }
    }
}




